program s
  use input_mdl, only: muUV,muIR,muMS,dp
  use thdm_ct, only: initct
  use fill_couplings, only: fill_couplings_mdl
  use lbfgsb_interface, only: chi2func_ptr,dchi2func_ptr,init_grid,set_bounds,run_loop
  use modelinput, only: linc,printp,rs_ref,rs_cmp
  use class_particles, only: print_parameters_mdl
  implicit none

  ! Options:
  ! enable linear conversion (rather than full conversion)
  !logical :: linc = .false.

  procedure(chi2func_ptr), pointer :: chi2f
  procedure(dchi2func_ptr), pointer :: dchi2f
  real(dp) :: ref_a0,ref_b0,a,b,chires,cab_in,tb_in
  real(dp), parameter :: sol_thr = 1d-15

  character(len=100) :: file_bp
  character(len=*), parameter :: fmt_s = '(2x,a)'
  character(len=*), parameter :: fmt = '(2x,a,x,f23.14)'
  character(len=*), parameter :: fmt_i = '(2x,a,x,i1)'
  character(len=*), parameter :: fmt_a = '(2x,a,x,a)'

  chi2f => chi2_thdm
  dchi2f => dchi2_thdm

  call read_input

  if (printp .gt. 0) call print_parameters_mdl(printp-1)

  if (linc .eq. 0) then
    call run_loop
    call write_result
  else
    call lin_conversion
    call write_linc_result
  end if

  contains  

  subroutine write_result
    use class_particles, only: tb,cab
    write(*,*)
    write(*,fmt_a) "Full conversion from " // trim(rs_ref) // " to " // trim(rs_cmp)

    call chi2_thdm([real(tb,kind=dp),real(cab,kind=dp)],chires)
    if (chires .gt. sol_thr) then
      write(*,*) "Only approximate solution."
    end if
    write(*,fmt) "    tb:", real(tb)
    write(*,fmt) "   cab:", real(cab)
    write(*,fmt) "chires:", chires
    write(*,*)
  end subroutine write_result

  subroutine write_linc_result
    use class_particles, only: tb,cab
    write(*,*)
    write(*,fmt_s) rs_cmp

    write(*,fmt_a) "Linear conversion from " // trim(rs_ref) // " to " // trim(rs_cmp)
    write(*,fmt) "    tb:", real(tb)
    write(*,fmt) "   cab:", real(cab)
    write(*,fmt) "chires:", 0.
    write(*,*)
  end subroutine write_linc_result

  subroutine read_input
    use fill_couplings, only: fill_couplings_mdl
    use thdm_ct, only: initct
    use collier_interface, only: init_collier
    use modelinput, only: read_modelinput,cab,tb,l5,MHH,MHA,MHC,printp,ytype,mumsbsm
    use modelfile, only: set_parameter_mdl
    real(dp) :: mhh_in,mha_in,mhc_in,l5_in,massreg
    integer :: ytype_in
    complex(dp),parameter :: cone=1d0,cnul=0d0

    call read_modelinput

    muUV = 100d0
    muIR = 100d0
    muMS = 100d0

    mhh_in = MHH
    mha_in = MHA
    mhc_in = MHC
    cab_in = cab
    tb_in = tb
    l5_in = l5
    ytype_in = ytype

    if (printp .eq. 0) then
      write(*,fmt) "mhh in:", mhh_in
      write(*,fmt) "mha in:", mha_in
      write(*,fmt) "mhc in:", mhc_in
      write(*,fmt) "cab in:", cab_in
      write(*,fmt) "tb  in:", tb_in
      write(*,fmt) "l5  in:", l5_in
      write(*,fmt_i) "ytype in:", ytype_in
      write(*,fmt_a) "rs ref:", trim(rs_ref)
      write(*,fmt_a) "rs cmp:", trim(rs_cmp)
    end if

    call set_tb_cab_rcl(tb_in,cab_in)
    call set_parameter_mdl('l5',  l5_in*cone)
    call set_Z2_thdm_yukawa_type_rcl(ytype_in)
    call set_parameter_mdl('MHH', mhh_in*cone)
    call set_parameter_mdl("WHH", cnul)
    call set_parameter_mdl('MHC', mhc_in*cone)
    call set_parameter_mdl("WHC", cnul)
    call set_parameter_mdl('MHA', mha_in*cone)
    call set_parameter_mdl("WHA", cnul)


    if (mumsbsm .eq. -1d0) then
      muMS = (125d0 + mhh_in + mha_in + 2*mhc_in)/5d0
    else
      muMS = mumsbsm
    end if
    muUV = muMS
    call set_parameter_mdl('muMS_BSM', muMS*cone)

    call init_collier
    call fill_couplings_mdl()
    call initct(0d0)
    call set_ref_values
    call init_grid(2,10)
    call set_initial_bounds

  end subroutine read_input

  subroutine set_initial_bounds()
    call set_bounds([0d0, -1d0],     &
                    [100d0, 1d0],    &
                    [tb_in, cab_in], &
                    chi2f,dchi2f)
    
  end subroutine set_initial_bounds

  subroutine set_tb_cab_rcl(tb,cab)
    use class_particles, only: pi,cb,sa
    real(dp), intent(in) :: tb,cab
    real(dp)             :: b,a

    b = atan(tb)
    cb = cos(b)
    a = acos(cab) + b

    if (a .gt. pi/2d0) then
      a = 2*b - a
    end if
    sa = sin(a)
  end subroutine set_tb_cab_rcl

  subroutine set_Z2_thdm_yukawa_type_rcl (ytype)
    use modelfile, only: set_parameter_mdl
    complex(dp),parameter :: cone=1d0,cnul=0d0

    ! This subroutine sets the yukawa type in the softly-broken Z2 symmetric
    ! THDM to "ytype".
    ! ytype = 1 : Type-I   h1u=h1d=h1l=0
    ! ytype = 2 : Type-II  h1u=h2d=h2l=0
    ! ytype = 3 : Type-X   h1u=h1d=h2l=0
    ! ytype = 4 : Type-Y   h1u=h2d=h1l=0

    integer, intent(in) :: ytype
    select case (ytype)
    case (1)
      call set_parameter_mdl("h1u", cnul)
      call set_parameter_mdl("h2u", cone)

      call set_parameter_mdl("h1d", cnul)
      call set_parameter_mdl("h2d", cone)

      call set_parameter_mdl("h1l", cnul)
      call set_parameter_mdl("h2l", cone)
    case (2)
      call set_parameter_mdl("h1u", cnul)
      call set_parameter_mdl("h2u", cone)

      call set_parameter_mdl("h2d", cnul)
      call set_parameter_mdl("h1d", cone)

      call set_parameter_mdl("h2l", cnul)
      call set_parameter_mdl("h1l", cone)
    case (3)
      call set_parameter_mdl("h1u", cnul)
      call set_parameter_mdl("h2u", cone)

      call set_parameter_mdl("h1d", cnul)
      call set_parameter_mdl("h2d", cone)

      call set_parameter_mdl("h2l", cnul)
      call set_parameter_mdl("h1l", cone)
    case (4)
      call set_parameter_mdl("h1u", cnul)
      call set_parameter_mdl("h2u", cone)

      call set_parameter_mdl("h2d", cnul)
      call set_parameter_mdl("h1d", cone)

      call set_parameter_mdl("h1l", cnul)
      call set_parameter_mdl("h2l", cone)
    end select

  end subroutine set_Z2_thdm_yukawa_type_rcl

  subroutine lin_conversion
    use class_particles, only: tb,cab,pi,cb,sa
    real(dp)             :: b,a,an,bn,tbr,cabr

    tbr=tb
    b = atan(tbr)
    cb = cos(b)
    cabr=cab
    a = acos(cabr) + b

    if (a .gt. pi/2d0) then
      a = 2*b - a
    end if
    an = a + get_da(rs_ref) - get_da(rs_cmp)
    bn = b + get_db(rs_ref) - get_db(rs_cmp)
    sa = sin(an)
    cab = cos(an-bn)
    tb = tan(bn)
    chires = 0
  end subroutine lin_conversion

  function get_da(scheme)

    use thdm_ct, only: da_FJTS,da_MDTS,da_onshell_bfm,da_OS_T1,da_OS_T2, &
                       da_l3_MS_FJTS,da_l3_MS_MDTS
    character(len=*) :: scheme
    real(dp)         :: get_da

    select case (scheme)
      case ('FJ')
        get_da = da_FJTS
      case ('PR')
        get_da = da_MDTS
      case ('l3_MDTS')
        get_da = da_l3_MS_MDTS
      case ('l3_FJTS')
        get_da = da_l3_MS_FJTS
      case ('BFM')
        get_da = da_onshell_bfm
      case ('OS1')
        get_da = da_OS_T1
      case ('OS2')
        get_da = da_OS_T2
      case ('OS12')
        get_da = da_OS_T2
      case default
        write(*,*) "Requested RS: ",scheme
        write(*,*) "but not available."
        write(*,*) "Allowed schemes: FJ, PR, OS1, OS2, OS12, BFM"
        stop
        
    end select 
    
  end function get_da

  function get_db(scheme)
    use thdm_ct, only: db_FJTS,db_MDTS,db_BFM,db_BFM_OD,db_OS_T1,db_OS_T2,db_OS_T1T2
    character(len=*) :: scheme
    real(dp)         :: get_db

    select case (scheme)
      case ('FJ')
        get_db = db_FJTS
      case ('PR')
        get_db = db_MDTS
      case ('l3_MDTS')
        get_db = db_MDTS
      case ('l3_FJTS')
        get_db = db_FJTS
      case ('BFM')
        get_db = db_BFM
      case ('OS1')
        get_db = db_OS_T1
      case ('OS2')
        get_db = db_OS_T2
      case ('OS12')
        get_db = db_OS_T1T2
      case default
        write(*,*) "Requested RS: ",scheme
        write(*,*) "but not available."
        write(*,*) "Allowed schemes: FJ, PR, OS1, OS2, OS12, BFM"
        stop
        
    end select 
    
  end function get_db

  subroutine set_ref_values
    use class_particles, only: tb,cab,pi
    real(dp) :: b,a,tbr,cabr

    tbr=tb
    b = atan(tbr)
    cabr=cab
    a = acos(cabr) + b

    if (a .gt. pi/2d0) then
      a = 2*b - a
    end if
    ref_a0 = a + get_da(rs_ref)
    ref_b0 = b + get_db(rs_ref)
  end subroutine set_ref_values

  subroutine chi2_thdm(x,res)
    use thdm_ct, only: initct
    use class_particles, only: tb,cab,pi,init_parameters_mdl
    use fill_couplings, only: fill_couplings_mdl
    real(kind=kind(23d0)), dimension(:), intent(in)  :: x
    real(kind=kind(23d0)),               intent(out) :: res
    real(dp)             :: b,a,tbr,cabr

    call set_tb_cab_rcl(x(1),x(2))
    call init_parameters_mdl
    call fill_couplings_mdl()
    call initct(0d0)

    tbr=tb
    b = atan(tbr)
    cabr=cab
    a = acos(cabr) + b

    if (a .gt. pi/2d0) then
      a = 2*b - a
    end if

    res = (ref_b0 - b - get_db(rs_cmp))**2 + (ref_a0 - a - get_da(rs_cmp))**2
    
  end subroutine chi2_thdm

  subroutine dchi2_thdm(dim,x,df)

    integer,                  intent(in)  :: dim
    real(dp), dimension(dim), intent(in)  :: x 
    real(dp), dimension(dim), intent(out) :: df
    real(dp) :: fval,yp,ym
    real(dp), parameter :: eps=1d-8
    real(dp), dimension(dim) :: x_tmp
    integer :: i

    call chi2_thdm(x,fval)
    x_tmp = x
    do i = 1, dim
      x_tmp(i) = x(i)+eps
      call chi2_thdm(x_tmp,yp)
      x_tmp(i) = x(i)-eps
      call chi2_thdm(x_tmp,ym)
      df(i) = (yp-ym)/2d0/eps
      x_tmp(i) = x(i)
    end do
   
  end subroutine dchi2_thdm

end program s
