module pde_integrator 

  abstract interface
    subroutine update_func(yc)
    complex(kind=kind(23d0)), dimension(:), optional, intent(in) :: yc
    end subroutine update_func

    subroutine set_kernel (x,f)
    real(kind=kind(23d0)),                  intent(in)  :: x
    complex(kind=kind(23d0)), dimension(:), intent(out) :: f
    end subroutine set_kernel
  end interface

  contains
  subroutine euler_step(y,x,stepsize,reset_func,set_k)
    implicit none
    complex(kind=kind(23d0)), dimension(2), intent(inout) :: y
    real(kind=kind(23d0)),                  intent(inout) :: x
    real(kind=kind(23d0)),                  intent(in)    :: stepsize
    complex(kind=kind(23d0)), dimension(2)                :: f
    procedure (update_func), pointer, intent(in)          :: reset_func
    procedure (set_kernel), pointer, intent(in)        :: set_k

    ! compute f(y(x),x)
    call set_k(x,f)

    x = x+stepsize
    y(:) = y(:) + f(:)*stepsize

  end subroutine euler_step

  subroutine rk4_step(y,x,stepsize,dim,reset_func,set_k)
    complex(kind=kind(23d0)), dimension(:), intent(inout) :: y
    real(kind=kind(23d0)),                  intent(inout) :: x
    real(kind=kind(23d0)),                  intent(in)    :: stepsize
    integer,                                intent(in)    :: dim
    procedure (update_func), pointer, intent(in)          :: reset_func
    procedure (set_kernel), pointer, intent(in)        :: set_k
    complex(kind=kind(23d0)), dimension(dim)              :: k1,k2,k3,k4,yc
    ! set k1 = f(y(x), x)
    call set_k(x,k1)
    ! update y(x)-> y(x)+k1/2
    yc(:) = y(:) + stepsize*k1(:)/2d0
    call reset_func(yc)

    ! set k2 = f(y(x)+k1/2,x+h/2)
    call set_k(x+stepsize/2d0,k2)
    ! update y(x) -> y(x)+k2/2
    yc(:) = y(:) + stepsize*k2(:)/2d0
    call reset_func(yc)

    ! set k3 = f(y(x)+k2/2,x+h/2)
    call set_k(x+stepsize/2d0,k3)
    ! update y(x) -> y(x)+k3
    yc(:) = y(:) + stepsize*k3(:)
    call reset_func(yc)

    ! set k4 = f(y(x)+k3,x+h)
    call set_k(x+stepsize,k4)

    ! final step: y += (k1 + 2k2 + 2k3 + k4)/6
    y(:) = y(:) + stepsize*(k1(:) + 2*k2(:) + 2*k3(:) + k4(:))/6d0
    x = x + stepsize

  end subroutine rk4_step

  subroutine rk45_step(y,x,stepsize,dim,reset_func,set_k)
    complex(kind=kind(23d0)), dimension(:), intent(inout) :: y
    real(kind=kind(23d0)),                  intent(inout) :: x
    real(kind=kind(23d0)),                  intent(in)    :: stepsize
    integer,                                intent(in)    :: dim
    procedure (update_func), pointer, intent(in)          :: reset_func
    procedure (set_kernel), pointer, intent(in)        :: set_k
    complex(kind=kind(23d0)), dimension(dim) :: k1,k2,k3,k4,k5,k6,yc,yc2
    real(kind=kind(23d0)) :: error, stps

    call set_k(x,k1)
    yc(:) = y(:) + stepsize*k1(:)/4d0
    call reset_func(yc)
    call set_k(x+stepsize/4d0,k2)

    yc(:) = y(:) + stepsize*(3*k1(:) + 9*k2(:))/32d0
    call reset_func(yc)
    call set_k(x+stepsize*3d0/8d0,k3)

    yc(:) = y(:) + stepsize*(1932*k1(:)-7200*k2(:)+7296*k3(:))/2197d0
    call reset_func(yc)
    call set_k(x+stepsize*12d0/13d0,k4)

    yc(:) = y(:) + stepsize*(439d0*k1(:)/216d0-8d0*k2(:)+ &
            3680d0*k3(:)/513d0-845d0*k4(:)/4104d0)
    call reset_func(yc)
    call set_k(x+stepsize,k5)

    yc(:) = y(:) + stepsize*(-8d0*k1(:)/27d0+2d0*k2(:) &
            -3544d0*k3(:)/2565d0+1859d0*k4(:)/4104d0-11d0*k5(:)/40d0)
    call reset_func(yc)
    call set_k(x+stepsize/2,k6)


    !rk4 step
    yc(:) = y(:) + stepsize*(25d0*k1(:)/216d0 + 1408d0*k3(:)/2565d0 + &
                             2197d0*k4(:)/4101d0 - k5(:)/5d0)
    !rk5 step
    yc2(:) = y(:) + stepsize*(16d0*k1(:)/135d0 + 6656d0*k3(:)/12825d0 + &
                              28561d0*k4(:)/56430d0 - 9*k5(:)/50d0 + 2d0*k6(:)/55d0)
    x = x + stepsize
    error = maxval(abs(yc2(:)-yc(:))/abs(yc2))
    !write(*,*) "error:", error
    !stps = sqrt(sqrt(1d-16*minval(abs(yc2))/error/2d0))*stepsize
    !if (abs(stps) .lt. abs(stepsize)) then
    !  y = yc2
    !end if
    !write(*,*) "stps:", stps
    !stepsize = stps
    !if (error .gt. 1d-15) then
    !  stepsize = stepsize/2d0
    !else
    !  stepsize = stepsize*2d0
    !end if
    y = yc2
    !else
    !  y = yc2
    !end if
    !write(*,*) "stepsize:", stepsize

  end subroutine rk45_step



end module pde_integrator
